#!/usr/bin/python
# -*- coding: utf-8 -*-

'''Created by Daniel Sindhikara, sindhikara@gmail.com

Placevent

This program is designed to automatically place explicit solvent atoms/ions
based on 3D-RISM data. The 3D-RISM correlations should be in a DX file.

The details of the algorithm are described in: 

Placevent: An algorithm for prediction of explicit solvent atom distribution --
Application to HIV-1 protease and F-ATP synthase 

Daniel J. Sindhikara, Norio Yoshida, Fumio Hirata
http://dansindhikara.com/Software/Entries/2012/6/22_Placevent_New.html

This package requires Numpy.

g(r)_0 is printed in the occupation column
g(r)_i is printed in the beta column

    Copyright (C) 2012 Daniel J. Sindhikara

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

TODO: 
Add bulk water to avoid evacuation errors
Switch to Argparse-based method
'''

from math import *
import sys
import numpy as np
import grid.grid as grid
import ppdb.pdb as pdb


def converttopop(distribution, delta, conc):
    '''Convert distribution fn to population fn using concentration.

    Returns: 
    1) Population function (numpy array)
    2) Expected total population in grid (float)
    3) Single voxel volume (float) 
    '''

    xlen = len(distribution)
    ylen = len(distribution[0])
    zlen = len(distribution[0][0])
    popzero = [[[0.0 for z in range(zlen)] for y in range(ylen)]
               for x in range(xlen)]
    gridvolume = delta[0] * delta[1] * delta[2]
    print '# Volume of each voxel is ', gridvolume, ' cubic angstroms'

    # z fast, y med, x slow

    totalpop = 0
    for i in range(xlen):
        for j in range(ylen):
            for k in range(zlen):
                popzero[i][j][k] = distribution[i][j][k] * conc \
                    * gridvolume
                totalpop += popzero[i][j][k]
    print '# The population is approximately ', totalpop, \
          '  within the given grid'
    return (popzero, totalpop, gridvolume)


def doplacement(
    popzero,
    conc,
    gridvolume,
    origin,
    delta,
    shellindices,
    grcutoff,
    ):
    '''Does iterative portion of placement according to Placevent Algorithm

    Returns:
    NumPy array of "Atom" Class objects
    '''

    popi = np.array(popzero)  # popi (popzero) are instantaneous (initial) pops.
    bulkvoxelpop = gridvolume * conc
    max = 0
    topindices = [[[]]]
    print '# Initial 3D-RISM g(r) will be printed in occupancy column'
    placedcenters = []
    finished = 0
    print '# Doing placement...'
    index = 0
    while finished == 0:
        index += 1
        mymax = popi.max()
        (maxi, maxj, maxk) = np.argwhere(popi == mymax)[0]
        topindices.append([maxi, maxj, maxk])
        remainder = 1  # Remaining population to subtract
        indexradius = 0  # First shell radius
        while remainder > 0:
            availablepopulation = 0
            try:
                for indices in shellindices[indexradius]:  # Loop on shell.
                    try:
                        availablepopulation += popi[maxi + indices[0],
                                maxj + indices[1], maxk + indices[2]]
                    except IndexError:
                        availablepopulation += bulkvoxelpop # Steal from pseudobulk
                        #print '# Stopping! Population search went outside' \
                        #      ' grid.\n'
                        #print '# You may want to try again with a higher ' \
                        #      'concentration\n'
                        #remainder = -1  # Cue quitting search.
                        #finished = 1
                        #break
            except IndexError:
                print ' # Stopping!\n'
                print ' # Population search went outside maximum shell size.\n'
                print ' # This usually happens with a dilute concentration.\n'
                print ' # You may want to try again with a higher concentration.\n'
                print ' # Your results may still be reasonable.'
                remainder = -1  # Cue quitting search.
                finished = 1
                break
            if availablepopulation > remainder and remainder > -1:

                # There is enough in this shell to bring the remainder to zero.
                # Is this fraction right?
                fraction = (availablepopulation - remainder) / availablepopulation
                remainder = 0
                for indices in shellindices[indexradius]:
                    try:
                        popi[maxi + indices[0]][maxj + indices[1]][maxk
                            + indices[2]] = popi[maxi
                            + indices[0]][maxj + indices[1]][maxk
                            + indices[2]] * fraction
                    except IndexError:
                        pass # Don't need to adjust population outside grid
            elif remainder > -1:
                # Not enough remainder in this shell, set whole shell pop to zero.

                for indices in shellindices[indexradius]:
                    try:
                        popi[maxi + indices[0]][maxj + indices[1]][maxk
                            + indices[2]] = 0
                    except IndexError:
                        pass # Don't need to adjust population outside grid
                remainder -= availablepopulation
                indexradius += 1
                if indexradius > len(shellindices):  # Search is larger than maxshells
                    print ' # Stopping!\n'
                    print ' # Population search went outside grid.\n'
                    print ' # You may want to try again with a higher ' \
                          '  concentration.'
                    exit()

        myg0 = popzero[maxi][maxj][maxk] / (conc * gridvolume)  # Original g(r).
        mygi = mymax / (conc * gridvolume)  # g(r) when placed.
        placedcenters.append(pdb.Atom(serial=len(placedcenters),
                             resseq=index, coord=[maxi * delta[0]
                             + origin[0], maxj * delta[1] + origin[1],
                             maxk * delta[2] + origin[2]], occ=myg0,
                             tfac=mygi))
        if myg0 < grcutoff:
            finished = 1  # Already placed all atoms > grcutoff.
    return placedcenters


def returncenters(guvfilename, molar, grcutoff):
    '''Given distribution, returns placed atoms

    Input: guvfilename and molarity
    Returns: Placed centers as Atom objects
    '''

    conc = molar * 6.0221415E-4
    shellindices = grid.readshellindices()
    if guvfilename[-3:] == ".dx":
        (distribution, origin, delta, gridcount) = \
            grid.readdx(guvfilename)
    elif guvfilename[-3:] == ".h5":
        grids = grid.h5ToGrids(guvfilename)
        print "Warning, assuming target molecule is 'O', please contact the developer for more info."
        if 'guv' not in grids['O'].keys():
            exit("Error, guv distribution not found in file")
        gGrid = grids['O']['guv']
        distribution = gGrid.distribution
        origin = gGrid.origin
        delta = gGrid.deltas
        gridcount = gGrid.gridcount
    else:
        exit("Error, incompatible file type! -> {0}".format(guvfilename))

    (popzero, totalpop, gridvolume) = converttopop(distribution,
            delta, conc)
    return doplacement(popzero, conc, gridvolume, origin, delta, shellindices,
        grcutoff)


def main():
    print '# placevent.py <.dx or MDF .h5 file> <concentration M (molar)> [cutoff g(r) ' \
          ' (default 1.5)]\n'
    print '# Concentration (#/A^3) ~= molarity * 6.0221415E-4'
    if len(sys.argv) < 3:
        print 'Insufficient arguments need 2 : ', len(sys.argv) - 1
        exit()
    if len(sys.argv) == 4:
        grcutoff = float(sys.argv[3])
    else:
        grcutoff = 1.5
    guvfilename = sys.argv[1]
    molar = float(sys.argv[2])
    print '# Your .dx file is', guvfilename, 'molarity is', molar, 'M'
    placedcenters = returncenters(guvfilename, molar, grcutoff)
    for center in placedcenters:
        print str(center)[:-2]  # [:-2] is to get rid of the '\n'


if __name__ == '__main__':
    main()



'''
    Dan Sindhikara sindhikara@gmail.com
    Does PDB related stuff
    Todo: 
1) Integrate better with AMBER modules (e.g. guessatom, atom radii, etc)



'''
import math

class Atom():
    '''
Atom object
Stores atomic information as would be in a pdb line.

Below is PDB format
1-6 7-11 13-16 17 18-20 22 23-26 27 31-38 39-46 47-54 55-60 61-66 77-78 79-80
Record name Integer Atom Character Residue name Character Integer AChar Real(8.3) Real(8.3) Real(8.3) Real(6.2) Real(6.2) LString(2) element LString(2) charge
"ATOM " serial name altLoc resName chainID resSeq iCode
Atom serial number. Atom name. Alternate location indicator. Residue name. Chain identifier. Residue sequence number. Code for insertion of residues. Orthogonal coordinates for X in Angstroms. Orthogonal coordinates for Y in Angstroms. Orthogonal coordinates for Z in Angstroms. Occupancy. Temperature factor. Element symbol, right-justified. Charge on the atom.
    '''
    def __init__(self,serial=1,name="D",altloc=" ",resname="BCD",chainid=1,resseq=1,icode=" ",coord=[0.0,0.0,0.0],occ=1.0,tfac=0.0,element="D",charge=0.0,hetatm=False,ister=False):
        self.name = name[0:4]
        self.serial=serial
        self.altloc=altloc[0]
        self.resname = resname[0:3]
        self.chainid=str(chainid)[0]
        self.resseq=resseq
        self.icode=icode
        self.coord=coord
        self.occ=occ
        self.tfac=tfac
        self.element=element[0]
        self.charge=charge
        self.hetatm=hetatm
        self.ister=ister # this is the last Atom of the unit. (the next line in the pdb will be say "TER"
    def __str__(self):
        '''
    Returns a PDB format string
        '''
        if not self.hetatm :
            pdbstring="ATOM  %5d %4s%1s%3s %1c%4d%1s   %8.3f%8.3f%8.3f%6.2f%6.2f           %1s%2d\n" % (self.serial,self.name,self.altloc,self.resname,self.chainid,self.resseq,self.icode,self.coord[0],self.coord[1],self.coord[2],self.occ,self.tfac,self.element,self.charge)
        else :
            pdbstring="HETATM%5d %4s%1s%3s %1c%4d%1s   %8.3f%8.3f%8.3f%6.2f%6.2f           %1s%2d\n" % (self.serial,self.name,self.altloc,self.resname,self.chainid,self.resseq,self.icode,self.coord[0],self.coord[1],self.coord[2],self.occ,self.tfac,self.element,self.charge)
        if self.ister :
            pdbstring="%sTER\n" % pdbstring
        return(pdbstring) 

    def makestring(self):
        '''
    Antiquated function same as __str__
        '''
        print "# Warning, antiquated function, __str__ function added"
        return(str(self))

        return(pdbstring)
        
def centermolecule(listofAtoms):
    '''
Given a list of Atom class objects, returns the same list centered
    '''
    center=[0.0,0.0,0.0]
    for atom in listofAtoms:
        for dim in range(3):
            center[dim]+=atom.coord[dim]
    for dim in range(3):
        center[dim]=center[dim]/float(len(listofAtoms))
    for i in range(len(listofAtoms)):
        for dim in range(3):
            listofAtoms[i].coord[dim]-=center[dim]
    return(listofAtoms)


def readcoordinates(pdbfilename):
    ''' Given a pdbfilename, return list of coordinates (2dlist)
This function is less functional than readpdbintoAtoms()
It only stores coordinates
    '''
    pdbfile=open(pdbfilename)
    coords=[]
    for line in pdbfile:
        ident=line[0:6]
        if ident == "ATOM  " or ident == "HETATM" :
            coords.append([float(line[30:38]),float(line[38:46]),float(line[46:54])])
    return(coords)

def readpdbintoAtoms(pdblines,guesselement=False,guesshetatm=False):
    ''' Given a pdb file of arbitrary length parsed into lines,
store the data as a list of "Atom" class objects
    '''

    molecule=[]
    knownelements = ["C","N","O","P","H","S","K","KA"] # put larger ones at the end (so that KA supersedes K)
    knownresidues = ["THR","ILE","ALA","ARG","HIS","HIE","HID","CYS","CYX","LYS","ASP","ASH","GLU","SER","THR","ASN","THR","ASN","GLN","GLY","PRO","LEU","MET","PHE","TRP","TYR","VAL"]
    # The above lists are used to help classify atoms
    if guesselement :
        pass
        #print "Warning! Guesselement only uses single letter atoms"
    terminilist=[]
    for i,line in enumerate(pdblines):
        ident=line[0:6]
        if ident == "ATOM  " or ident == "HETATM" :
            #index in line is one less than column number
            serial=int(line[6:11])
            name=line[12:16]
            altloc=line[16]
            resname=line[17:20]
            chainid=line[21]
            resseq=int(line[22:26])
            icode=line[26]
            coord=[float(line[30:38]),float(line[38:46]),float(line[46:54])]
            occ=float(line[54:60])
            tfac=float(line[60:66])
            if len(line) >= 77 :
                element=line[76:78]
            else:
                element="  "
            if guesselement :
                for knownelement in knownelements:
                    if knownelement in name:
                        element=knownelement
            if len(line) >= 79:
                try :
                    charge=float(line[78:80])
                except :
                    charge=0.0
            else:
                charge=0.0
            hetatm=False
            if ident == "HETATM" :
                hetatm=True
            elif (resname not in knownresidues) and guesshetatm:
                hetatm=True
            molecule.append(Atom(serial,name,altloc,resname,chainid,resseq,icode,coord,occ,tfac,element,charge,hetatm))
        if line[0:3] == "TER" :
            molecule[-1].ister=True
    return(molecule)

def isitinside(coordinate,atoms,buffer=0.0):
    '''
Test if coordinate is inside VDW radius of list of atoms
    '''
    radii={'H' : 0.6, 'O' : 1.6612, 'C' : 1.908, 'S': 2.000, 'N' : 1.8240, 'KA' : 1.7131,  'C0' : 1.7131, 'P' : 2.1000, 'K' : 1.7131}
    inside = False
    for atom in atoms:
        if not atom.ister:
            dist=(coordinate[0]-atom.coord[0])**2+(coordinate[1]-atom.coord[1])**2+(coordinate[2]-atom.coord[2])**2
            dist=math.sqrt(dist)
        
            try:
                 if dist < radii[atom.element[0]]+buffer :
                     inside = True
                     break
            except:
                print "Cannot find radius of atom named: ",atom.element,atom.name
                exit()
    return(inside)

def isitnear(coordinate,atoms,radius):
    '''
Test if coordinate is within specified radius of listed atoms
    '''
    inside = False
    for atom in atoms:
        dist=(coordinate[0]-atom.coord[0])**2+(coordinate[1]-atom.coord[1])**2+(coordinate[2]-atom.coord[2])**2
        dist=math.sqrt(dist)
        if dist < radius:
            inside = True
            break
    return(inside)


def calcmolvolume(atoms,gridspacing,probesize=0.0):
    import modules.djsgrid as grid
    print "This function is under testing and quite slow. This should be replaced by faster, more reliable code"

    volume=0.0
    print "spacing =",gridspacing
    gridvolume=gridspacing**3
    print "grid volume is ",gridvolume
    maxes=[-1000]*3
    mins=[1000]*3
    for atom in atoms:
        if atom.coord[0] < mins[0] :
            mins[0] = atom.coord[0]
        elif atom.coord[0] > maxes[0] :
            maxes[0] = atom.coord[0]
        if atom.coord[1] < mins[1] :
            mins[1] = atom.coord[1]
        elif atom.coord[1] > maxes[1] :
            maxes[1] = atom.coord[1]
        if atom.coord[2] < mins[2] :
            mins[2] = atom.coord[2]
        elif atom.coord[2] > maxes[2] :
            maxes[2] = atom.coord[2]
    # add some buffer
    for i in range(3):
        mins[i]-=5.0
        maxes[i]+=5.0

    spans=[]
    for i in range(3):
        spans.append(int((maxes[i]-mins[i])/gridspacing))

    radmap={'H' : 0.6, 'O' : 1.6612, 'C' : 1.908, 'S': 2.000, 'N' : 1.8240, 'KA' : 1.7131,  'C0' : 1.7131, 'P' : 2.1000, 'K' : 1.7131}
    shelldict={} # construct dictionary of shell indices (these contain all grid points radially from atom center which count as inside the atom
    print " don't need to compute shellindices more than once, do it for a big one and assign shells"
    allshellindices=grid.precomputeshellindices(int((2.5+probesize)/gridspacing))
    for radkey in radmap: 
        myrad=radmap[radkey]+probesize # effective radius
        maxshellindex=int(myrad/gridspacing+0.5) # largest shell index that counts as inside
        shelldict[radkey]=allshellindices[:maxshellindex]

    occupiedflags=[[[False for i in range(spans[0])] for j in range(spans[1])] for k in range(spans[2])]
    testvol=0
    for shellindex in shelldict["C"]:
        for shellpoint in shellindex:
            testvol+=gridvolume
    print "at this resolution, Carbon volume (plus probe) is",testvol
    print "by geometry it's", (4.0/3.0)*3.14159 * (radmap["C"]+probesize) **3

    occupiedpdbfile=open("dumb.pdb","w")
    testedpdbfile=open("tested.pdb","w")
    for atom in atoms:
        myindices=grid.getindicesfromcoord([atom.coord[0],atom.coord[1],atom.coord[2]],mins,[gridspacing]*3)
        #print atom.coord[0],atom.coord[1],atom.coord[2]
        #print grid.getcoordfromindices(myindices,mins,[gridspacing]*3)
        myshells=shelldict[atom.element]
        for shell in myshells: # 0th, 1st, 2nd shell, etc
            for shellindex in shell: # points that lay on a shell
                netindex=[]
                for i in range(3): # indices of each point on shell
                    #print "shellindex,myindex[i]=",shellindex[i],myindices[i]
                    netindex.append(shellindex[i]+myindices[i]) # origin of atom + shell index
                    #print spans
                #print "netindices=",netindex
                #print grid.getcoordfromindices(netindex,mins,[gridspacing]*3)
                coord=grid.getcoordfromindices(netindex,mins,[gridspacing]*3)
                testedpdbfile.write(Atom(coord=coord).makestring())
                if not occupiedflags[netindex[0]][netindex[1]][netindex[2]]:
                    volume+=gridvolume
                    occupiedflags[netindex[0]][netindex[1]][netindex[2]]=True
                    occupiedpdbfile.write(Atom(coord=coord).makestring())
    occupiedpdbfile.close()
    testedpdbfile.close()
    return(volume)







try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

config = {
    'description': 'Placevent solvent placement software',
    'author': 'Daniel J. Sindhikara',
    'url':'http://dansindhikara.com/Software/Entries/2012/6/22_Placevent_New.html',
    'download_url': 'http://dansindhikara.com/Software/Entries/2012/6/22_Placevent_New.html',
    'author_email': 'sindhikara@gmail.com',
    'version': '1.4',
    'install_requires': ['numpy','grid'],
    'packages': ['placevent', 'placevent.tests', 'ppdb'],
    'package_data': {'placevent.tests': ['1L2Y/diffs','1L2Y/*.*', '1L2Y/ref/*'],
                     },
    'license': 'GPL',
    'scripts': ['placevent/placevent.py', 'placevent/tests/1L2Y/placevent_test.sh'],
    'name': 'placevent'
}
setup(**config)  
